// This file is not for the faint of heart. But it can be significantly
// improved by those daring to.

#include "jengine.h"
#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <ltdl.h>
#include <cassert>
#include <sstream>
#include <map>

const bool TRACEMEM=false;

// typedef jarray::B B;
// typedef jarray::C C;
// typedef jarray::S S;
// typedef jarray::I I;
// typedef jarray::D D;
// typedef jarray::Z Z;

// --------- J typedefs.
// The below inclusion of a portion of J interface is needed to access
// the ttop variable of J instance to be able to cooperate with J
// garbage collection. In principle this all can be replaced by a single
// numeric offset to ttop, but that will have to be calculated for all
// possible platforms and architectures. By including the J typedefs we
// try to ensure the proper access to ttop on all architectures, which J
// supports.
#include "js.h"

#if SY_WINCE                                                                                
#define NFCALL          100L            /* wince     named fn call depth   */               
#define NFDEP           200L            /* wince           fn call depth   */               
#endif                                                                                      
#if SYS & SYS_MACOSX                                                                        
#define NFCALL           9000L          /* darwin    named fn call depth   */               
#define NFDEP           18000L          /* darwin          fn call depth   */               
#endif                                                                                      
#ifndef NFCALL                                                                              
#define NFCALL          10000L          /* all other named fn call depth   */               
#define NFDEP           20000L          /* all other       fn call depth   */               
#endif                                                                                      

#define NALP            256             /* size of alphabet                */               
#define NETX            2000            /* size of error display buffer    */               
#define NPATH           1024            /* max length for path names,      */               
                                        /* including trailing 0 byte       */
#define MLEN            (SY_64?63:31)

#include "jtype.h"
#include "jt.h"
// --------- end of J typedefs

typedef void* (_stdcall *JInitType)     ();
typedef int   (_stdcall *JDoType)       (J, C*);
typedef C*    (_stdcall *JGetLocaleType)(J);
typedef void  (_stdcall *JSMType)       (J, void*);
typedef void  (_stdcall *JFreeType)     (J);
typedef int   (_stdcall *JGetMType)     (J jt, C* name, I* jtype, 
					 I* jrank, I* jshape, I* jdata);
typedef int   (_stdcall *JSetMType)     (J jt, C* name, I* jtype,
					 I* jrank, I* jshape, I* jdata);

typedef void* (_stdcall *jtgaType)      (J jt, const I t, const I n, const I r, const I* s);
typedef void  (_stdcall *jtfrType)      (J jt, void* A);
typedef void* (_stdcall *jtgcType)      (J jt, void* w, I old);
typedef void* (_stdcall *jttpushType)   (J jt, void* w);
typedef I     (_stdcall *jttpopType)    (J jt, I oldtop);

std::ostream *out=NULL;

void _stdcall Joutput(J jt,I type, C* s) {
  // output type                                                                  const I MTYOFM  = 1;       /* formatted result array output */             
  const I MTYOER  = 2;       /* error output */                              
  const I MTYOLOG = 3;       /* output log */                                
  const I MTYOSYS = 4;       /* system assertion failure */                  
  const I MTYOEXIT = 5;      /* exit */                                      
  const I MTYOFILE= 6;       /* output 1!:2[2 */

  if (MTYOEXIT==type) { // J wants us to exit
    exit((int)(I)s);
  };
  (*out) << s << std::flush;
};

class jengine::Impl {
public:
  static JInitType pJInit;
  static JSMType pJSM;
  static JDoType pJDo;  
  static JGetMType pJGetM;
  static JSetMType pJSetM;
  static JFreeType pJFree;
  static jtgaType pjtga;
  static jtfrType pjtfr;
  static jtgcType pjtgc;
  static jttpushType pjttpush;  
  static jttpopType pjttpop;  
  static lt_dlhandle libj;
  int err;

  static void initJlibrary(std::ostream& o) {
    if (!libj) {
      out=&o;
      int error = lt_dlinit ()     /* Initialise libltdl. */
	|| lt_dladdsearchdir (".")    /* set the search path. */
	|| lt_dladdsearchdir ("./thirdparty/j")  
	|| lt_dladdsearchdir ("../thirdparty/j")
#ifdef JDLLDIR
	|| lt_dladdsearchdir (JDLLDIR)
#endif
	|| lt_dladdsearchdir ("../../thirdparty/j") /* this makes "make distcheck" work */
	|| !((libj=lt_dlopenext("libj"))||(libj=lt_dlopenext("j"))); /* load */
      
      if (error) {
	fprintf(stderr, "Fatal error loading J engine: %s\n", lt_dlerror ());
	exit(9);
      };
      // ok, we loaded the engine, now set the entry points
      error = !(
		(pJInit=(JInitType)lt_dlsym(libj, "JInit")) &&
		(pJSM=(JSMType)lt_dlsym(libj, "JSM")) &&
		(pJDo=(JDoType)lt_dlsym(libj, "JDo")) &&
		(pJGetM=(JGetMType)lt_dlsym(libj, "JGetM")) &&
		(pJSetM=(JSetMType)lt_dlsym(libj, "JSetM")) &&
		(pJFree=(JFreeType)lt_dlsym(libj, "JFree")) &&
		(pjtga=(jtgaType)lt_dlsym(libj, "jtga")) &&
		(pjtfr=(jtfrType)lt_dlsym(libj, "jtfr")) &&
		(pjtgc=(jtgcType)lt_dlsym(libj, "jtgc")) &&
		(pjttpush=(jttpushType)lt_dlsym(libj, "jttpush")) &&
		(pjttpop=(jttpopType)lt_dlsym(libj, "jttpop")) 
		);
      if (error) {
	fprintf(stderr, "Fatal error resolving J symbols: %s\n", lt_dlerror ());
	fprintf(stderr, "pJInit = %p\npJSM = %p\npJDo = %p\npJGetM = %p\n"
		"pSetA = %p\npJFree = %p\n",pJInit, pJSM, pJDo,pJGetM,
		pJSetM,pJFree);
	exit(9);
      };
    }
  }

  // Pointer to J instance. Even though it is a class variable, only one 
  // instance of J engine can be active at a time. This is because J does
  // not allow callbacks to function with multiple instances. This limitation
  // may be lifted in future (jengine code is ready for this).
  J jt;

  // pointer back to jengine, owning this ::Impl
  jengine *je;

  // callbacks into J engine for calling adverb verb arguments
  // monadically and dyadically.
  typedef int (_stdcall *verbcallback) ();               // 0 arg
  typedef int (_stdcall *monadcallback) (int y);         // 1 arg
  typedef int (_stdcall *dyadcallback)  (int x, int y);  // 2 args
 
  // because these variables are static, only a single J instance will
  // properly support user-defined adverbs. While this can be done differently
  // here, note that J itself does not support multiple-instance callbacks.
  static verbcallback verbcb;   // return the top of verbstack as text
  static monadcallback monadcb; // execute top of verbstack monadically
  static dyadcallback dyadcb;   // execute top of verbstack dyadically
  
  bool doJ(const std::string s) {
    I err=pJDo(jt, const_cast<char *>(s.c_str()));
    return err==0;
  };
  
  void *GA(const I t, const I n, const I r, const I* s) {
    return pjtga(jt,t,n,r,s);
  };

  void FR(void* hdr) {
    pjtfr(jt,hdr);
  };

  jarray::I PROLOG() {
    return jt->tbase+jt->ttop;
  };
  jarray::I EPILOG(jarray::I oldtop) {
    return pjttpop(jt,oldtop);
  };
  void* EPILOG(jarray::I oldtop, void *reshdr) {
    return pjtgc(jt, reshdr, oldtop);
  };

  // retrieves a named array from J
  const jarray get(const std::string name) {
    I jtype, jrank;
    I* jshape;
    void* jdata;
    err=pJGetM(jt, const_cast<char *>(name.c_str()), 
	       &jtype, &jrank, (I*)&jshape, (I *)&jdata);
    if (!err)  {
      // what we now actually need is just the beginning of the header,
      // which is always 7 words before the shape array.
      jarray res(je, jshape-7);
      return res;
    } else {
      jarray res;
      return res;
    };
  };

  static std::ostream & print_J_pointer(std::ostream &o, const jarray::I iptr) {
    if (iptr<0) 
      o<<'_'<<(-iptr);
    else
      o<<iptr;
    return o;
  };

  // assigns J name to the specified array
  bool set(const std::string name, jarray &value) {
    using namespace std;
    // this would not copy the array data, although parsing is required.
    ostringstream str;
    str << name << "=:" << "(15!:7) ";
    print_J_pointer(str, value.getHeader());
    return doJ(str.str());
    // // this will implicitly copy data.
    // I type=value.type();
    // I rank=value.rank();
    // I* shape=value.shape();
    // I* data=value.data();
    // err=pJSetM(jt, const_cast<char *>(name.c_str()), 
    // 	       &type, &rank, (I*)&shape, (I*)&data);
    // return err==0;
  };

  void stackHighlight(I addr, I oldttop) {
    std::cerr << '[';
    std::cerr << addr;
    std::cerr << ',';
    for (int i=0; i<jt->ttop; i++) {
      if ((oldttop>0)&&(oldttop==jt->tbase+i))
	std::cerr<<'|';
      if (((I)jt->tstack[i])==addr) std::cerr<<'*';
    };
    std::cerr << ']';
  };

  // this gets called from J, yielding us: the pointer to (this) jengine
  // the address of the c++ monad to call and the address of J symbol for
  // the argument of the monad.
  typedef struct {
    void* name;
    void* value;
  } Jsymbol;
  static jarray::I monadadaptor(jengine *this_, monad mf, Jsymbol* ysymb) {
    jarray y(this_, ysymb->value);
    if (TRACEMEM) {
      std::cerr << std::endl;
      std::cerr << "- y"; this_->pImpl->stackHighlight((I)ysymb->value,0);
      std::cerr << " refcount of y was at start " << y.getRefcount()-1 << std::endl;
    };    
    jarray::I _ttop=this_->PROLOG();
    jarray r=mf(y);
    assert(r.getEngine()==this_->pImpl->je);  // the array we pass back to J,
                                              // should be allocated by this 
                                              // jengine instance
    I rhdr=r.getHeader();
    this_->EPILOG(_ttop, (void*)rhdr);
    if (TRACEMEM) {
      std::cerr << "- OUT"; this_->pImpl->stackHighlight(rhdr, _ttop);
      std::cerr << " refcount at OUT is now " << r.getRefcount() << std::endl;
    };
    return rhdr;
  };
  static jarray::I dyadadaptor(jengine *this_, dyad df, Jsymbol* xsymb, Jsymbol* ysymb) {
    jarray x(this_, xsymb->value);
    if (TRACEMEM) {
      std::cerr << std::endl;
      std::cerr << "-- x"; this_->pImpl->stackHighlight((I)xsymb->value,0);
      std::cerr << " refcount of x was at start " << x.getRefcount()-1 << std::endl;
    };
    jarray y(this_, ysymb->value);
    if (TRACEMEM) {
      std::cerr << "-- y"; this_->pImpl->stackHighlight((I)ysymb->value,0);
      std::cerr << " refcount of y was at start " << y.getRefcount()-1 << std::endl;
    };
    I _ttop=this_->PROLOG();
    jarray r=df(x,y);
    assert(r.getEngine()==this_->pImpl->je);  // the array we pass back to J,
                                              // should be allocated by this 
                                              // jengine instance
    I rhdr=r.getHeader();
    this_->EPILOG(_ttop, (void*)rhdr);
    if (TRACEMEM) {
      std::cerr << "-- OUT"; this_->pImpl->stackHighlight(rhdr, _ttop);
      std::cerr << " refcount at OUT is now " << r.getRefcount() << std::endl;
    };
    return rhdr;
  };
  
  // scripts, defined in J+
  static std::map<std::string,std::string> builtins;
  bool isBuiltin(std::string name) const {
    return builtins.count(name)>0;
  };
  std::set<std::string> getBuiltins() const {
    std::set<std::string> res;
    std::map<std::string,std::string>::iterator it;
    for(it=builtins.begin();it!=builtins.end(); it++)
      res.insert((*it).first);
    return res;
  };
  static jarray topverbasstring(jengine *je) {
    using namespace std;
    // if the argument array is invalid, which should never happen
    // normally, return the verb textual representation
    jarray verb=jarray(je, (void *)verbcb());
    verb.release(); // now we own this memory
    // J does a very good service, fixing a verb ! But, in our case
    // the string version would contain memory addresses of various
    // callbacks. This defies its usefullness for hashing.
    // Here the returned string is scanned and all the stringified
    // scripts are replaced back by their names.
    string str;
    verb.get(str);
    map<string,string>::iterator it;
    for (it=builtins.begin() ; it != builtins.end(); it++ ) {
      string name=(*it).first;
      string text=(*it).second;
      // replace all occurences of text by name
      int i = 0;
      while((i = str.find(text, i)) != string::npos) {
	str.replace(i, text.length(), name);
	i+=name.length();
      };
    };
    jarray nverb(je, str);
    return nverb;
  };
  // these dispatch the calls to J callback
  static jarray monadcbadaptor(jarray y) {
    jengine *je_=y.getEngine();
    if (TRACEMEM) {
      std::cerr << std::endl;
      std::cerr << "+ refcount of y, "<<",";
      je_->pImpl->stackHighlight(y.getHeader(false), 0);
      std::cerr<<" at cb " << y.getRefcount() << std::endl;    
    };
    I _ttop=je_->PROLOG();
    if (TRACEMEM) std::cerr << "+ttop=" << _ttop << std::endl;
    void* reshdr=(void *)monadcb(y.getHeader(false));
    jarray ra=jarray(je_, reshdr);
    pjttpush(je_->pImpl->jt, reshdr);
    if (TRACEMEM) {
      std::cerr << "+ OUT"; 
      je_->pImpl->stackHighlight((I)reshdr, _ttop); 
      std::cerr << std::endl;
    };
    return ra;
  };
  static jarray dyadcbadaptor(jarray x, jarray y) {
    if (!y.isValid()) return topverbasstring(x.getEngine());
    assert(x.getEngine()==y.getEngine());
    jengine *je_=y.getEngine();
    if (TRACEMEM) {
      std::cerr << std::endl;
      std::cerr << "++ refcount of x, "<<",";
      je_->pImpl->stackHighlight(x.getHeader(false), 0);
      std::cerr<<" at cb " << x.getRefcount() << std::endl;    
      std::cerr << "++ refcount of y, "<<",";
      je_->pImpl->stackHighlight(y.getHeader(false), 0);
      std::cerr<<" at cb " << y.getRefcount() << std::endl;    
    };
    I _ttop=je_->PROLOG(); // PROLOG
    if (TRACEMEM) std::cerr << "+ttop=" << _ttop << std::endl;
    void* reshdr=(void *)dyadcb(x.getHeader(false), y.getHeader(false));
    jarray ra=jarray(je_, reshdr);
    pjttpush(je_->pImpl->jt, reshdr);
    if (TRACEMEM) {
      std::cerr << "++ OUT"; 
      je_->pImpl->stackHighlight((I)reshdr, _ttop); 
      std::cerr << std::endl;
    };
    return ra;
  };
  // these pass the callback adaptors to the user-defined adverb code
  static jarray::I amonadadaptor(jengine *this_, amonad mf, Jsymbol* ysymb) {
    jarray y(this_, ysymb->value);
    if (TRACEMEM) {
      std::cerr << std::endl;
      std::cerr << "= y"; 
      this_->pImpl->stackHighlight((I)ysymb->value,0);
      std::cerr << " refcount of y was at start " << y.getRefcount()-1 << std::endl;
    };
    I _ttop=this_->PROLOG();
    jarray r=mf(&monadcbadaptor, &dyadcbadaptor, y);
    assert(r.getEngine()==this_->pImpl->je);  // the array we pass back to J,
                                              // should be allocated by this 
                                              // jengine instance
    I reshdr=r.getHeader();
    this_->EPILOG(_ttop, (void*)reshdr);
    if (TRACEMEM) {
      std::cerr << "= OUT"; 
      this_->pImpl->stackHighlight(reshdr, _ttop);
      std::cerr << " refcount at OUT is now " << r.getRefcount() << std::endl;
    };
    return reshdr;
  };
  static jarray::I adyadadaptor(jengine *this_, adyad df, Jsymbol* xsymb, Jsymbol* ysymb) {
    jarray x(this_, xsymb->value);
    if (TRACEMEM) {
      std::cerr << std::endl;
      std::cerr << "== x"; 
      this_->pImpl->stackHighlight((I)xsymb->value,0);
      std::cerr << " refcount of x was at start " << x.getRefcount()-1 << std::endl;
    };
    jarray y(this_, ysymb->value);
    if (TRACEMEM) {
      std::cerr << "== y"; 
      this_->pImpl->stackHighlight((I)ysymb->value,0);
      std::cerr << " refcount of y was at start " << y.getRefcount()-1 << std::endl;
    };
    I _ttop=this_->PROLOG();    
    jarray r=df(&monadcbadaptor, &dyadcbadaptor, x,y);
    assert(r.getEngine()==this_->pImpl->je);  // the array we pass back to J,
                                              // should be allocated by this 
                                              // jengine instance
    I reshdr=r.getHeader();
    this_->EPILOG(_ttop, (void*)reshdr);
    //pjttpush(jt, (void*)reshdr);
    if (TRACEMEM) {
      std::cerr << "== OUT"; 
      this_->pImpl->stackHighlight(reshdr, _ttop);
      std::cerr << " refcount at OUT is now " << r.getRefcount() << std::endl;
    };
    return reshdr;
  };

  // defines multiline named script
  bool defScript(std::string name, int type, std::string code, 
		 int mr=jengine::RMAX, 
		 int lr=jengine::RMAX, 
		 int rr=jengine::RMAX) {
    using namespace std;
    // since doJ only executes a single line, script needs to be
    // converted to character codes.
    ostringstream str;
    str << name << "=: " << type << " : (";
    for (int i=0; i<code.size(); i++) str << " " << (int)code[i];
    str << "{ a.)";
    // ranks:
    // by default, explicit verbs have all infinite ranks
    if ((mr!=jengine::RMAX)||(lr!=jengine::RMAX)||(rr!=jengine::RMAX)) {
      str << " \" ";
      if ((lr==rr)&&(mr==lr)) { // all ranks are the same, use short form
	str << mr; // here mr is never infinite
      } else {
	if (mr!=rr) {  // "lr rr is a shorthand for  "rr lr rr
	  if (mr==jengine::RMAX) str << "_"; else str << mr;
	  str << ' ';
	};
	if (lr==jengine::RMAX) str << "_"; else str << lr;
	str << ' ';
	if (rr==jengine::RMAX) str << "_"; else str << rr;
      };
    };
    bool defined=doJ(str.str());
    if (defined) { 
      // here we read back from J and store the fixed version
      // of the script to be able to reduce it back to its
      // name, when the textual version of the verb is requested
      doJ("temporaryname=: '"+name+"' f.");
      doJ("temporaryname=: (5!:5) <'temporaryname'");
      jarray tn=get("temporaryname");
      assert(tn.type()==jarray::T_LIT);
      string text;
      tn.get(text);
      //cerr << endl << name << "=: " << text << endl;
      builtins.insert(pair<string,string>(name,text));
    };
    return defined;
  };

  bool defVerbAdverb(std::string name, jarray::I mf, jarray::I df, bool adverb,
		     int mr=jengine::RMAX, 
		     int lr=jengine::RMAX, 
		     int rr=jengine::RMAX) {
    using namespace std;
    string script;
    {
      ostringstream str;
      if (mf) {
	// monadic case
	// str << "output=. 2&((1!:2)~)" << endl;
	str << "ay =. y" << endl; // black magic, but otherwise 'y' is undefined
	str << "(4!:55) < 'ay'" << endl;
	if (adverb)   // adverb: push verb argument to callback stack
	  str << "verbstack =: (verbstack ` u)" << endl << "r =. ";
	str << "(15!:7) '0 ";
	int adaptor=adverb?(jarray::I)&amonadadaptor:(jarray::I)&monadadaptor;
	print_J_pointer(str,adaptor);
	str << " > x x x x' (15!:0) ( ";
	print_J_pointer(str, (jarray::I)je);
	str << ';';
	print_J_pointer(str,mf);
	str << " ; (15!:6) <'y' )" << endl;
	if (adverb)   // adverb: pop verb argument from callback stack
	  str << "verbstack =: }: verbstack " << endl << "r" << endl;
	if (df) str << ":" << endl;
      };
      if (df) {
	// dyadic case
        // str << "output=. 2&((1!:2)~)" << endl;
	if (adverb)   // adverb: push verb argument to callback stack
	  str << "verbstack =: (verbstack ` u)" << endl << "r =. ";
	str << "(15!:7) '0 ";
	int adaptor=adverb?(jarray::I)&adyadadaptor:(jarray::I)&dyadadaptor;
	print_J_pointer(str,adaptor);
	str << " > x x x x x' (15!:0) ( ";
	print_J_pointer(str, (jarray::I)je);
	str << ';';
	print_J_pointer(str,df);
	str << " ; ((15!:6) <'x') ; (15!:6) <'y' )" << endl;
	if (adverb)   // adverb: pop verb argument from callback stack
	  str << "verbstack =: }: verbstack" << endl << "r" << endl;
      };
      script=str.str();
    }
    return defScript(name, (adverb?1:(!mf?4:3)), script, mr, lr, rr);
  };

  bool defVerb(std::string name, monad mf, dyad df,
	       int mr, int lr, int rr) {
    return defVerbAdverb(name, (jarray::I) mf, (jarray::I) df, false, 
			 mr, lr, rr);
  };

  bool defAdverb(std::string name, amonad mf, adyad df) {
    return defVerbAdverb(name, (jarray::I) mf, (jarray::I) df, true);
  };

  Impl(jengine *je_):jt(NULL),je(je_) {
    using namespace std;
    assert(libj&&pJInit&&pJSM&&pJDo&&pJGetM&&pJSetM&&pJFree);

    jt=(J)pJInit();
    if (jt==0) {
	fprintf(stderr, "Fatal error initializing J engine.");
	exit(9);
    }
    
    // smoptions
    const I SMWIN = 0;  /* j.exe    Jwdw (Windows) front end */
    const I SMJAVA = 2; /* j.jar    Jwdp (Java) front end */
    const I SMCON = 3;  /* jconsole */

    void* callbacks[] = {(void*)Joutput,0,/* Jinput */ 0,0,(void*)SMCON};
    pJSM(jt, callbacks);

    err=0;

    // get addresses of callbacks
    {
      doJ("cb =: (15!:13) 0");
      jarray cb=get("cb");
      assert(cb.type()==jarray::T_INT);
      assert(cb.size()==1);
      int cbint;
      cb.get(cbint,0);
      verbcb=(verbcallback)cbint;
    }

    {
      doJ("cb =: (15!:13) 1");
      jarray cb=get("cb");
      assert(cb.type()==jarray::T_INT);
      assert(cb.size()==1);
      int cbint;
      cb.get(cbint,0);
      monadcb=(monadcallback)cbint;
    }

    {
      doJ("cb =: (15!:13) 2");
      jarray cb=get("cb");
      assert(cb.type()==jarray::T_INT);
      assert(cb.size()==1);
      int cbint;
      cb.get(cbint,0);
      dyadcb=(dyadcallback)cbint;
    }
    doJ("verbstack =: (4!:55) < 'cb'"); // clean up, erase 'cb'
    doJ("verbstack =: ''"); // empty array to serve as verbstack
    
    // define the callback dispatcher in J
    // it takes the verb from the top of verbstack and executes it
    // on given argument arrays monadically or dyadically, as requested.
    string cb(
	      // "output=. 2&((1!:2)~)" "\n"
	      "na =. # y" "\n"
	      "au =. {: verbstack" "\n"
	      "if. (na=1) do." "\n" // monadic call
	      "r =. (au @. 0) (15!:7) y" "\n"
	      "elseif. (na=2) do." "\n"          // dyadic call
	      "r =. ((15!:7) {. y) (au @. 0) ((15!:7) {: y)" "\n"
	      "elseif. (na=0) do." "\n" // verb on top of the stack as text
	      "auf =. ((au @. 0) f.)" "\n"
	      "r =. (5!:5) < 'auf'" "\n"
	      "end." "\n"
	      "ar =. {. (15!:1) ((15!:6) < 'r'),4,1,4" "\n"
	      "rc =. (15!:1) ar,16,1,4" "\n" // refcount
	      "(rc+1) (15!:2) ar,16,1,4" "\n"// artificially inc by 1 to
	                                     // prevent J from freeng the memory
	      "ar" "\n"
	      );
    bool ok=defScript("cdcallback", 3, cb);
    assert(ok);
  }

  ~Impl() {
    pJFree(jt);
    jt=NULL;
  };

private:
  // copy disabled
  Impl(const Impl &);
  const Impl& operator=(const Impl&);
};


JInitType jengine::Impl::pJInit=NULL;
JSMType jengine::Impl::pJSM=NULL;
JDoType jengine::Impl::pJDo=NULL;  
JGetMType jengine::Impl::pJGetM=NULL;
JSetMType jengine::Impl::pJSetM=NULL;
JFreeType jengine::Impl::pJFree=NULL;  
jtgaType jengine::Impl::pjtga=NULL;  
jtfrType jengine::Impl::pjtfr=NULL;  
jtgcType jengine::Impl::pjtgc=NULL;  
jttpushType jengine::Impl::pjttpush=NULL;  
jttpopType jengine::Impl::pjttpop=NULL;  
lt_dlhandle jengine::Impl::libj=NULL;
jengine::Impl::verbcallback jengine::Impl::verbcb=NULL;
jengine::Impl::monadcallback jengine::Impl::monadcb=NULL;
jengine::Impl::dyadcallback jengine::Impl::dyadcb=NULL;

std::map<std::string, std::string> jengine::Impl::builtins;

jengine::jengine():
  pImpl(new jengine::Impl(this)) {
};

jengine::~jengine() {
  delete pImpl;
};

void jengine::initJlibrary(std::ostream & o) {
  jengine::Impl::initJlibrary(o);
}

bool jengine::doJ(const std::string s) {
  return pImpl->doJ(s);
};

// retrieves a named array from J
const jarray jengine::get(const std::string name) {
  return pImpl->get(name);
};

// assigns J name to the specified array
bool jengine::set(const std::string name, jarray &value) {
  return pImpl->set(name,value);
};

// returns J error code for the last unsuccessful operation.
int jengine::getError() {
  return pImpl->err;
};

bool jengine::defVerb(std::string name, monad mf, dyad df,
		      int mr, int lr, int rr) {
  return pImpl->defVerb(name, mf, df, mr, lr, rr);
};

bool jengine::defScript(std::string name, int type, std::string code, 
			int mr, int lr, int rr) {
  return pImpl->defScript(name, type, code, mr, lr, rr);
};

bool jengine::defAdverb(std::string name, amonad mf, adyad df) {
  return pImpl->defAdverb(name, mf, df);
};

bool jengine::isBuiltin(std::string name) const {
  return pImpl->isBuiltin(name);
};

std::set<std::string> jengine::getBuiltins() const {
  return pImpl->getBuiltins();
};

void* jengine::GA(const jarray::I t, const jarray::I n, 
		  const jarray::I r, const jarray::I *s) {
  return pImpl->GA(t,n,r,s);
};
void jengine::FR(void *hdr) {
  pImpl->FR(hdr);
};
jarray::I jengine::PROLOG() {
  return pImpl->PROLOG();
};
jarray::I jengine::EPILOG(jarray::I oldtop) {
  return pImpl->EPILOG(oldtop);
};
void* jengine::EPILOG(jarray::I oldtop, void* reshdr) {
  return pImpl->EPILOG(oldtop, reshdr);
};
