#include "config.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>
#include "trjfile.h"

// this should be ported
#include <unistd.h>

using namespace std;

int main (int argc, char **argv) {
  bool error=false;
  bool opt_h=false;
  bool opt_t=false;
  int c;

  while (((c = getopt (argc, argv, "ht")) != -1)&&!error)
    switch (c) {
    case 'h':
      opt_h = true;
      break;
    case 't':
      opt_t = true;
      break;
    case '?':
      error=true;
    default:
      abort ();
    }
  if (!(error)&&(error=(opt_h&&opt_t))) {
    cerr << "Only one of -h or -t may be specified at the same time." << endl;
  };
  if (!(error)&&(error=(!(opt_h||opt_t)))) {
    cerr << "Some action needs to be specified via -h or -t options." << endl;
  };
  if ((!error)&&(error=((!opt_h)&&(argc!=optind+1)))) {
    cerr << "There must be exactly one non-option argument -- "
      "trajectory file name."<<endl;
  };
  if (error) {
    cerr << "Please use -h option to see help or check the man page." << endl;
    exit(9);
  };
  if (opt_h) {
    char help[] = {
#include "yacts-trj_help.h"
    };
    cout << "Welcome to yacts-trj : yacts trajectory file utility." << endl 
	 << endl << help 
	 << "For additional details see the man page." << endl;
    exit(0);
  };
  if (opt_t) {
    trjfile trj_in;
    trj_in.open(argv[optind],false);
    if (!trj_in.good()) {
      cerr << "Error opening input trajectory file \""
	   << argv[optind] << "\"." << endl;
      exit(8);
    };
    int neq;
    if ((neq=trj_in.getNEQ())<=0) {
      cerr << "Trajectory file is corrupt or incompletely formed."<<endl; 
      exit(8);
    };
    if (trj_in.size()<1) {
      cerr << "Trajectory must contain at least one frame."<<endl;
      exit(8);
    };
    if (trj_in.size()==1) {
      cerr << "Trajectory already contains only one frame, nothing to do."<<endl;
      exit(0);
    };
    char *tname=strdup("trajectiry.tmp.XXXXXX");
    if (mkstemp(tname)<0) {
      cerr << "Error creating temporary file." << endl;
      exit(8);
    };
    trjfile trj_out;
    trj_out.open(tname, true);
    double *data;
    double T;
    data = (double*) malloc(neq*sizeof(double));
    trj_in.toFrame(-1); // go to the last frame
    trj_in.loadFrame(T,data);
    trj_out.saveFrame(neq, T, data);
    trj_in.close();
    trj_out.close();
    if (remove(argv[optind])!=0) {
      cerr << "Error removing input trajectory file \""<<
	argv[optind]<<"\""<<endl;
    };
    if (rename(tname, argv[optind])!=0) {
      cerr << "Error renaming the created trajectory file \""<<
	argv[optind]<<"\""<<endl;
    };
    free(data);
    free(tname);
  };
};
